/* Hobbit's efforts at a unified s/key. */

#include "generic.h"

#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#else
#include <sys/types.h>
#endif
#include <string.h>

#ifdef	__MSDOS__
#include <dos.h>
#else	/* Assume unix; "TERMIOS" fix is implicit below */
#include <fcntl.h>
#include <signal.h>
/* #endif -- but wait, there's more */

/* SYSV terminalisms swiped from Chasin's version */

#undef USE_ECHO			/* default to not echoing! */

#ifdef stty
# undef stty
#endif
#ifdef gtty
# undef gtty
#endif

#ifndef SYSV
# include <sgtty.h>
# define TTYSTRUCT sgttyb
# define stty(fd,buf) ioctl((fd),TIOCSETN,(buf))
# define gtty(fd,buf) ioctl((fd),TIOCGETP,(buf))
#else
# include <termio.h>
# define TTYSTRUCT termio
# define stty(fd,buf) ioctl((fd),TCSETA,(buf))
# define gtty(fd,buf) ioctl((fd),TCGETA,(buf))
#endif

#ifdef SYSV
    struct termio newtty;
    struct termio oldtty;
#else
    struct sgttyb newtty;
    struct sgttyb oldtty;
    struct tchars chars;
#endif

#ifdef SIGVOID
#define SIGTYPE void
#else
#define SIGTYPE void
#endif

SIGTYPE trapped();	/* XXX */

#endif /* MSDOS */

/* NOTE:  All MD5 ifdef's were done by Dan McD.  and FIXED.  _H*/

#ifdef MD5
#include "md5.h"
#else
#include "md4.h"
#endif

#include "skey.h"

/* Crunch a key:
 * concatenate the seed and the password, run through MD4 and
 * collapse to 64 bits. This is defined as the user's starting key.
 */
int
keycrunch(result,seed,passwd)
char *result;	/* 8-byte result */
char *seed;	/* Seed, any length */
char *passwd;	/* Password, any length */
{
	char *buf;
#ifdef MD5
	MD5_CTX md;
#else
	MD4_CTX md;
#endif
	UINT4 results[4];
	unsigned int buflen;
	
	buflen = strlen(seed) + strlen(passwd);
	if((buf = (char *)malloc(buflen+1)) == NULL)
		return -1;
	strcpy(buf,seed);
	strcat(buf,passwd);

	/* Crunch the key through MD[45] */
	sevenbit(buf);
#ifdef DEBUG
printf ("in KC: buf is: /%s/ at %lx\n", buf, buf);
#endif
#ifdef MD5
	MD5Init(&md);
	MD5Update(&md,(unsigned char *)buf,buflen);
	MD5Final((unsigned char *)results,&md);
#else
	MD4Init(&md);
	MD4Update(&md,(unsigned char *)buf,buflen);
	MD4Final((unsigned char *)results,&md);
#endif
	free(buf);

#ifdef DEBUG		/* Can see the MDx hash digest here. */
printf ("result   block: %lx %lx %lx %lx siz %d\n",
  results[0], results[1], results[2], results[3], sizeof (results));
#endif
	/* Fold result from 128 to 64 bits */
	results[0] ^= results[2];
	results[1] ^= results[3];
#ifdef DEBUG
printf ("smooshed block: %lx %lx %lx %lx siz %d\n",
  results[0], results[1], results[2], results[3], sizeof (results));
#endif
	memcpy(result,(char *)results,8);
#ifdef DEBUG
printf ("memcpyed block starts: %2x %2x %2x %2x %2x...\n",
  result[0],result[1],result[2],result[3],result[4]);
#endif
	return 0;
}

/* The one-way function f(). Takes 8 bytes and returns 8 bytes in place */
void
f(x)
char *x;
{
#ifdef MD5
        MD5_CTX md;
#else
	MD4_CTX md;
#endif
	UINT4 results[4];
#ifdef DEBUG
printf ("in F: x is: /%s/ at %lx\n", x, x);
#endif

#ifdef MD5
	MD5Init(&md);
	MD5Update(&md,(unsigned char *)x,8);
	MD5Final((unsigned char *)results,&md);
#else
	MD4Init(&md);
	MD4Update(&md,(unsigned char *)x,8);
	MD4Final((unsigned char *)results,&md);
#endif
#ifdef DEBUG
printf ("result   block: %lx %lx %lx %lx siz %d\n",
  results[0], results[1], results[2], results[3], sizeof (results));
#endif
	/* Fold 128 to 64 bits */
	results[0] ^= results[2];
	results[1] ^= results[3];
#ifdef DEBUG
printf ("smooshed block: %lx %lx %lx %lx siz %d\n",
  results[0], results[1], results[2], results[3], sizeof (results));
#endif

	/* Only works on byte-addressed little-endian machines!! */
	memcpy(x,(char *)results,8);
#ifdef DEBUG
printf ("memcpyed block starts: %2x %2x %2x %2x %2x...\n",
  x[0],x[1],x[2],x[3],x[4]);
#endif
}

/* Strip trailing cr/lf from a line of text */
void
rip(buf)
char *buf;
{
	char *cp;

	if((cp = strchr(buf,'\r')) != NULL)
		*cp = '\0';

	if((cp = strchr(buf,'\n')) != NULL)
		*cp = '\0';
}

#ifdef	__MSDOS__
char *
readpass(buf,n)
char *buf;
int n;
{
	int i;
	char *cp;

	for(cp=buf,i = 0; i < n ; i++)
		if ((*cp++ = bdos(7,0,0)) == '\r')
			break;
	*cp = '\0';
	printf("\n");
	rip(buf);
	backspace(buf);		/* added.  _H*/
	return buf;
}

#else /* MSDOS */
char *
readpass(buf,n)
char *buf;
int n;
{
#ifdef OLD_NRL			/* xxx: berkeley specific, I bet */
	int fflags,lword,lwordsav;
	struct sgttyb ttyf,ttysave;

	/* Set normal line editing */
	fflags = fcntl(fileno(stdin),F_GETFL,0);
	fcntl(fileno(stdin),F_SETFL,fflags & ~FNDELAY);
	ioctl(fileno(stdin),TIOCLGET,&lword);
	ioctl(fileno(stdin),TIOCLGET,&lwordsav);
	lword |= LCRTERA|LCRTKIL;
	ioctl(fileno(stdin),TIOCLSET,&lword);

	/* Turn off echoing */
	ioctl(fileno(stdin), TIOCGETP, &ttyf);
	ioctl(fileno(stdin), TIOCGETP, &ttysave);
	ttyf.sg_flags &= ~(ECHO|RAW|CBREAK);
	ttyf.sg_flags |= CRMOD;
	ioctl(fileno(stdin),TIOCSETP,&ttyf);
	fgets(buf,n,stdin);
	rip(buf);

	/* Restore previous tty modes */
	fcntl(fileno(stdin),F_SETFL,fflags);
	ioctl(fileno(stdin),TIOCSETP,&ttysave);
	ioctl(fileno(stdin),TIOCLSET,&lwordsav);

#else /* OLD_NRL */

#ifndef USE_ECHO
    set_term ();
    echo_off ();
#endif

	fgets(buf,n,stdin);

	/*
	after the secret key is taken from the keyboard, the line feed is
	written to standard error instead of standard output.  That means that
	anyone using the program from a terminal won't notice, but capturing
	standard output will get the key words without a newline in front of
	them. 
	*/
        fprintf(stderr, "\n");
        fflush(stderr);
#ifndef USE_ECHO
    unset_term ();
#endif

	rip(buf);
	sevenbit(buf);
	backspace(buf);		/* added.  _H*/
	return buf;
#endif /* OLD_NRL */
}

/* XXX: perhaps merge Colin Campbell's TERMIOS stuff in here */
set_term ()
{
    gtty (fileno(stdin), &newtty);
    gtty (fileno(stdin), &oldtty);
 
    signal (SIGINT, trapped);
}

echo_off ()
{

#ifdef SYSV
    newtty.c_lflag &= ~(ICANON | ECHO | ECHONL);
#else
    newtty.sg_flags |= CBREAK;
    newtty.sg_flags &= ~ECHO;
#endif

#ifdef SYSV
    newtty.c_cc[VMIN] = 1;
    newtty.c_cc[VTIME] = 0;
    newtty.c_cc[VINTR] = 3;
#else
    ioctl(fileno(stdin), TIOCGETC, &chars);
    chars.t_intrc = 3;
    ioctl(fileno(stdin), TIOCSETC, &chars);
#endif

    stty (fileno (stdin), &newtty);
}

unset_term ()
{
    stty (fileno (stdin), &oldtty);
 
#ifndef SYSV
    ioctl(fileno(stdin), TIOCSETC, &chars);
#endif
}

void trapped()
 {
  signal (SIGINT, trapped);
  printf ("^C\n");
  unset_term ();
  exit (-1);
 }

#endif /* MSDOS */

/* removebackspaced over charaters from the string */
/* added DEL as backspace equivalent, and ^U as line kill.  _H*/
backspace(buf)
char *buf;
{
	char bs = 0x8;
	char del = 0x7f;
	char cu = 0x15;		/* ^U */
	char *cp = buf;
	char *out = buf;

	while(*cp) {
		if (( *cp == bs ) || ( *cp == del)) {
			if(out == buf) {
				cp++;
				continue;
			} else {
			  cp++;
			  out--;
			}
		} else {
			*out++ = *cp++;
		} /* bs or del */

		if (*cp == cu) {
			out = buf;
			cp++;
		}
	} /* while *cp */

	*out = '\0';
}

sevenbit(s)
char *s;
{
	/* make sure there are only 7 bit code in the line*/
	while(*s){
		*s = 0x7f & ( *s);
		s++;
	}
}
